﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Regular
{
    public class Automat
    {
        /// <summary>
        /// Seznam stavů automatu
        /// </summary>
        public List<Stav> Stavy { get; set; }

        /// <summary>
        /// Oznamuje zda již byla provedena reverze hran všech stavů automatu
        /// </summary>
        public bool ProvedenaReverzeHran { get; set; }

        public Automat()
        {
            this.Stavy = new List<Stav>();
            this.ProvedenaReverzeHran = false;
            this.PridejStav();
        }

        public void PridejStav()
        {
            this.Stavy.Add(new Stav());
        }

        /// <summary>
        /// Započne výpočet automatu
        /// </summary>
        /// <param name="vstupAutomatu">Vstup, který má být automatem porovnán</param>
        /// <param name="definovanZacatek">Byl definován začátek slova?</param>
        /// <param name="definovanKonec">Byl definován konec slova?</param>
        /// <returns></returns>
        public bool StartAutomatu(string vstupAutomatu, bool definovanZacatek, bool definovanKonec)
        {

            string vstup = vstupAutomatu;
            bool konec = false;

            if (definovanZacatek == true)
            {
                //nemažu po písmenech, ale všechny znaky až po mezeru včetně
                while (!(konec))
                {
                    konec = ProjdiAutomat(vstup, definovanKonec);
                    if (vstup.Count() > 0)
                    {
                        int mezera = vstup.IndexOf(' ');
                        
                        if (mezera == -1)
                        {
                            //ve výrazu není mezera
                            break;
                        }
                        else
                        {
                            vstup = vstup.Remove(0, mezera + 1);
                        }
                        
                    }
                    else
                    {
                        break;
                    }
                }

            } else
            {
                //odmazávám po jednom znaku
                while (!(konec))
                {
                    konec = ProjdiAutomat(vstup, definovanKonec);
                    if (vstup.Count() > 0)
                    {
                        vstup = vstup.Remove(0, 1);
                    } else
                    {
                        break;
                    }
                    
                }

            }

            //někdy se hodí odpálit automatické přechody jako první
            if (konec == false && this.ProvedenaReverzeHran == false)
            {
                foreach (Stav s in this.Stavy)
                {
                    s.Hrany.Reverse();
                }
                this.ProvedenaReverzeHran = true;
                konec = this.StartAutomatu(vstupAutomatu, definovanZacatek, definovanKonec);
            }

            return (konec);
        }

        /// <summary>
        /// Samotný výpočet automatu
        /// </summary>
        /// <param name="vstupAutomatu">Vstup automatu</param>
        /// <param name="definovanKonec">Byl definován konec?</param>
        /// <returns></returns>
        private bool ProjdiAutomat(string vstupAutomatu, bool definovanKonec)
        {
            string vstup = vstupAutomatu;
            int aktualniStav = 0;
            bool vstupAkceptovan = true;

            for (int z=0;z<=(vstup.Length-1);z++)
            {
                bool nalezenaHrana = false;
                foreach (Hrana h in this.Stavy[aktualniStav].Hrany)
                {
                    if (vstup[z] == h.Znak || h.Znak == '%')
                    {
                        aktualniStav = h.NasledujiciStav;
                        nalezenaHrana = true;
                        if (h.Znak == '%')
                        {
                            z--;
                        }
                        break;
                    }
                }

                if (this.Stavy[aktualniStav].KonecnyStav == true && definovanKonec == false)
                {
                    break;
                }

                if (nalezenaHrana == false)
                {   
                    vstupAkceptovan = false;
                    break;
                }
                
            }

            //existuje v poslednim stavu automatickej prechod?
            foreach (Hrana h in this.Stavy[aktualniStav].Hrany)
            {
                if (h.Znak == '%')
                {
                    aktualniStav = h.NasledujiciStav;
                }
            }


            if (vstupAkceptovan == true && this.Stavy[aktualniStav].KonecnyStav == true) 
            {
                return true;
            }
            else
            {
                return false;
            }

        }

    }
}
