﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Regular
{
    public class Regular
    {
        /// <summary>
        /// Regulární výraz
        /// </summary>
        public string RegVyraz { get; set; }

        /// <summary>
        /// Automat vygenerovaný pro zadaný regulární výraz
        /// </summary>
        public Automat AutomatVyrazu { get; set; }

        /// <summary>
        /// Byl definován začátek slova (^)
        /// </summary>
        public bool DefinovanZacatek { get; set; }

        /// <summary>
        /// Byl definován konec slova ($)
        /// </summary>
        public bool DefinovanKonec { get; set; }
        
        /// <summary>
        /// Výraz je validní
        /// </summary>
        public bool VyrazOK { get; set; }

        /// <summary>
        /// Vytvoří instanci Regular
        /// </summary>
        /// <param name="vyraz">Regulární výraz</param>
        public Regular(string vyraz)
        {
            this.DefinovanZacatek = false;
            this.DefinovanKonec = false;
            this.VyrazOK = false;
            this.RegVyraz = vyraz;
            this.VytvorStavy();
        }

        /// <summary>
        /// Na základě vstupu vrátí návratovou hodnotu regulárního výrazu
        /// </summary>
        /// <param name="vstup">Vstup, který chceme porovnat tímto regulárním výrazem</param>
        /// <returns>0=vstup prošel, 1=vstup neprošel, 2=regulární výraz není správně zapsaný</returns>
        public int VlozVstup(string vstup)
        {
            int navratovaHodnota = 0;

            if (this.VyrazOK == false)
            {
                navratovaHodnota = 2;
            } else
            {
                if (this.AutomatVyrazu.StartAutomatu(vstup, this.DefinovanZacatek, this.DefinovanKonec) == false)
                {
                    navratovaHodnota = 1;
                }
            }

            return navratovaHodnota;
        }

        /// <summary>
        /// Vytvoří automat pro daný regulární výraz (rovněž pomocí automatu)
        /// </summary>
        private void VytvorStavy()
        {
            this.AutomatVyrazu = new Automat();
            int stav = 1;
            string aktualniText = "";
            bool validniVyraz = true;
            foreach(char z in RegVyraz)
            {
                //v předchozím kroku for cyklu byl nějaký nepovolený znak
                if (validniVyraz == false)
                {
                    break;
                }

                switch (stav)
                {
                    case 1:
                        if (z == '^')
                        {
                            this.DefinovanZacatek = true;
                            stav = 2;
                        }
                        else if(z == '[')
                        {
                            stav = 3;
                        }
                        else if (z >= 48 && z<=57 || z>=65 && z<=90 || z>=97 && z<=122 || z == 32)
                        {
                            stav = 12;
                            aktualniText = z.ToString();
                        }
                        else
                        {
                            validniVyraz = false;
                        }
                        break;
                    case 2:
                        if (z >= 48 && z <= 57 || z >= 65 && z <= 90 || z >= 97 && z <= 122 || z == 32)
                        {
                            aktualniText = z.ToString();
                            stav = 12;
                        }
                        else if (z == '[')
                        {
                            stav = 3;
                        }
                        else
                        {
                            validniVyraz = false;
                        }
                        break;
                    case 3:
                        if (z == ']')
                        {
                            if (aktualniText == "")
                            {
                                validniVyraz = false;
                                break;
                            }
                            int poziceZobacku = aktualniText.IndexOf('^');
                            if (poziceZobacku >= 0)
                            {
                                string nezarazeneZnaky = aktualniText.Substring(poziceZobacku+1,aktualniText.Length-poziceZobacku-1);
                                aktualniText = aktualniText.Remove(poziceZobacku,nezarazeneZnaky.Length+1);
                                //velká písmena
                                for (int i = 65; i <= 90; i++)
                                {
                                    if (!(nezarazeneZnaky.Contains(Convert.ToChar(i))))
                                    {
                                        aktualniText += Convert.ToChar(i);
                                    }
                                }
                                //malá písmena
                                for (int i = 97; i <= 122; i++)
                                {
                                    if (!(nezarazeneZnaky.Contains(Convert.ToChar(i))))
                                    {
                                        aktualniText += Convert.ToChar(i);
                                    }
                                }
                                //číslice
                                for (int i = 48; i <= 57; i++)
                                {
                                    if (!(nezarazeneZnaky.Contains(Convert.ToChar(i))))
                                    {
                                        aktualniText += Convert.ToChar(i);
                                    }
                                }
                                if (!(nezarazeneZnaky.Contains(" ")))
                                {
                                    aktualniText += " ";
                                }
                            }
                            stav = 12;
                        }
                        else if (z == '^')
                        {
                            aktualniText += z.ToString();
                            stav = 10;
                        }
                        else if (z == '-')
                        {
                            if (aktualniText == "")
                            {
                                validniVyraz = false;
                            }
                            stav = 5;
                        }
                        else if (z == ':')
                        {

                            aktualniText += z.ToString();
                            stav = 4;
                        }
                        else if (z >= 48 && z <= 57 || z >= 65 && z <= 90 || z >= 97 && z <= 122 || z == 32)
                        {
                            
                            //stav zůstává 3
                            aktualniText += z.ToString();
                        }
                        else
                        {
                            validniVyraz = false;
                        }
                        break;
                    case 4:
                        if (z == ':')
                        {
                            int prvniDvojtecka = aktualniText.IndexOf(':');
                            //co bylo napsáno mezi dvojtečkami
                            string mnozina = aktualniText.Substring(prvniDvojtecka+1,aktualniText.Length - (prvniDvojtecka+1));
                            //odmažu text množiny a v podmínkách ho nahradim danými znaky
                            aktualniText = aktualniText.Remove(prvniDvojtecka, mnozina.Length + 1);
                            if (mnozina == "alpha")
                            {
                                //všechna písmena

                                //velká písmena
                                for (int i = 65; i<=90;i++)
                                {
                                    aktualniText += Convert.ToChar(i);
                                }
                                //malá písmena
                                for (int i = 97; i<=122;i++)
                                {
                                    aktualniText += Convert.ToChar(i);
                                }
                            }
                            else if(mnozina == "upper")
                            {
                                //velká písmena
                                for (int i = 65; i <= 90; i++)
                                {
                                    aktualniText += Convert.ToChar(i);
                                }
                            }
                            else if (mnozina == "lower")
                            {
                                //malá písmena
                                for (int i = 97; i <= 122; i++)
                                {
                                    aktualniText += Convert.ToChar(i);
                                }
                            }
                            else if (mnozina == "num")
                            {
                                //číslice
                                for (int i = 48; i <= 57; i++)
                                {
                                    aktualniText += Convert.ToChar(i);
                                }
                            }
                            else if (mnozina == "alnum")
                            {
                                //všechny znaky kromě mezery

                                //velká písmena
                                for (int i = 65; i <= 90; i++)
                                {
                                    aktualniText += Convert.ToChar(i);
                                }
                                //malá písmena
                                for (int i = 97; i <= 122; i++)
                                {
                                    aktualniText += Convert.ToChar(i);
                                }
                                //číslice
                                for (int i = 48; i <= 57; i++)
                                {
                                    aktualniText += Convert.ToChar(i);
                                }
                            }
                            else
                            {
                                validniVyraz = false;
                            }
                            stav = 3;
                        }
                        else if (z >= 48 && z <= 57 || z >= 65 && z <= 90 || z >= 97 && z <= 122)
                        {
                            aktualniText += z.ToString();
                        }
                        else
                        {
                            validniVyraz = false;
                        }
                        break;
                    case 5:
                        if (z >= 48 && z <= 57 || z >= 65 && z <= 90 || z >= 97 && z <= 122)
                        {
                            //zjistim počátek a konec intervalu písmen
                            int cisloKoncovehoZnaku = z;
                            int cisloPocatecnihoZnaku = aktualniText[aktualniText.Length - 1];
                            //pokud je první znak číslice tak i druhý musí být číslice (vyšší)
                            if (JeCisloZIntervalu(cisloPocatecnihoZnaku, 48, 57))
                            {
                                if (!(JeCisloZIntervalu(cisloKoncovehoZnaku,cisloPocatecnihoZnaku,57)))
                                {
                                    validniVyraz = false;
                                }
                            }
                            //pokud je první znak velké písmeno musí být i druhý velké písmeno (dál v abecedě)
                            else if (JeCisloZIntervalu(cisloPocatecnihoZnaku, 65, 90))
                            {
                                if (!(JeCisloZIntervalu(cisloKoncovehoZnaku, cisloPocatecnihoZnaku, 90)))
                                {
                                    validniVyraz = false;
                                }
                            }
                            //pokud je první znak malé písmeno musí být i druhý malé písmeno (dál v abecedě)
                            else if (JeCisloZIntervalu(cisloPocatecnihoZnaku, 97, 122))
                            {
                                if (!(JeCisloZIntervalu(cisloKoncovehoZnaku, cisloPocatecnihoZnaku, 122)))
                                {
                                    validniVyraz = false;
                                }
                            }
                            for (int i = cisloPocatecnihoZnaku+1;i<=cisloKoncovehoZnaku;i++)
                            {
                                aktualniText += Convert.ToChar(i);
                            }
                            stav = 3;
                        }
                        else
                        {
                            validniVyraz = false;
                        }
                        break;
                    case 6:
                        if (z >= 48 && z <= 57 || z >= 65 && z <= 90 || z >= 97 && z <= 122 || z == 32)
                        {
                            stav = 12;
                            aktualniText = z.ToString();
                        }
                        else if (z == '[')
                        {
                            stav = 3;
                        }
                        else if (z == '$')
                        {
                            this.DefinovanKonec = true;
                            stav = 9;
                        }
                        else
                        {
                            validniVyraz = false;
                        }
                        break;
                    case 7:
                        if (z >= 48 && z <= 57 || z >= 65 && z <= 90 || z >= 97 && z <= 122 || z == 32)
                        {
                            stav = 12;
                            aktualniText = z.ToString();
                        }
                        else if (z == '[')
                        {
                            stav = 3;
                        }
                        else if (z == '$')
                        {
                            this.DefinovanKonec = true;
                            stav = 9;
                        }
                        else
                        {
                            validniVyraz = false;
                        }
                        break;
                    case 8:
                        if (z >= 48 && z <= 57 || z >= 65 && z <= 90 || z >= 97 && z <= 122 || z == 32)
                        {

                            stav = 12;
                            aktualniText = z.ToString();
                        }
                        else if (z == '[')
                        {
                            stav = 3;
                        }
                        else if (z == '$')
                        {
                            this.DefinovanKonec = true;
                            stav = 9;
                        }
                        else
                        {
                            validniVyraz = false;
                        }
                        break;
                    case 9:
                        validniVyraz = false;
                        break;
                    case 10:
                        if (z >= 48 && z <= 57 || z >= 65 && z <= 90 || z >= 97 && z <= 122 || z == 32)
                        {
                            aktualniText += z.ToString();
                            stav = 3;
                        }
                        else
                        {
                            validniVyraz = false;
                        }
                        break;
                    //kde je 11?? :D
                    case 12:
                        if (z >= 48 && z <= 57 || z >= 65 && z <= 90 || z >= 97 && z <= 122 || z == 32)
                        {
                            int cisloPoslednihoStavu = this.AutomatVyrazu.Stavy.Count - 1;
                            this.AutomatVyrazu.PridejStav();
                            foreach (char znak in aktualniText)
                            {
                                this.AutomatVyrazu.Stavy[cisloPoslednihoStavu].PridejHranu(
                                    Convert.ToChar(znak), cisloPoslednihoStavu+1);
                            }

                            aktualniText = z.ToString();
                            //stav zůstává na 11
                        }
                        else if (z == '[')
                        {

                            int cisloPoslednihoStavu = this.AutomatVyrazu.Stavy.Count - 1;
                            this.AutomatVyrazu.PridejStav();
                            foreach (char znak in aktualniText)
                            {
                                this.AutomatVyrazu.Stavy[cisloPoslednihoStavu].PridejHranu(
                                    Convert.ToChar(znak), cisloPoslednihoStavu + 1);
                            }

                            aktualniText = "";
                            stav = 3;
                        }
                        else if (z == '*')
                        {
                            int cisloPoslednihoStavu = this.AutomatVyrazu.Stavy.Count - 1;
                            //hrany do stejného stavu (O-n opakování znaku)
                            foreach (char znak in aktualniText)
                            {
                                this.AutomatVyrazu.Stavy[cisloPoslednihoStavu].PridejHranu(
                                    Convert.ToChar(znak), cisloPoslednihoStavu);
                            }

                            aktualniText = "";
                            stav = 6;
                        }
                        else if (z == '+')
                        {
                            int cisloPoslednihoStavu = this.AutomatVyrazu.Stavy.Count - 1;
                            //jeden nový stav a pak hrany do stejného stavu (1-n opakování znaku)
                            this.AutomatVyrazu.PridejStav();
                            foreach (char znak in aktualniText)
                            {
                                this.AutomatVyrazu.Stavy[cisloPoslednihoStavu].PridejHranu(
                                    Convert.ToChar(znak), cisloPoslednihoStavu + 1);
                                this.AutomatVyrazu.Stavy[cisloPoslednihoStavu + 1].PridejHranu(
                                    Convert.ToChar(znak), cisloPoslednihoStavu + 1);
                            }

                            aktualniText = "";
                            stav = 7;
                        }
                        else if (z == '?')
                        {
                            int cisloPoslednihoStavu = this.AutomatVyrazu.Stavy.Count - 1;
                            //jeden nový stav (0-1 opakování znaku)
                            this.AutomatVyrazu.PridejStav();
                            foreach (char znak in aktualniText)
                            {
                                this.AutomatVyrazu.Stavy[cisloPoslednihoStavu].PridejHranu(
                                    Convert.ToChar(znak), cisloPoslednihoStavu + 1);
                            }
                            //automatický přechod na následující stav (ať už byl jednou nebo nulakrát)
                            //oboum posledním stavům přidám lambda přechod (%) na další stav
                            this.AutomatVyrazu.PridejStav();
                            this.AutomatVyrazu.Stavy[cisloPoslednihoStavu].PridejHranu(
                                '%', cisloPoslednihoStavu + 2);
                            this.AutomatVyrazu.Stavy[cisloPoslednihoStavu + 1].PridejHranu(
                                '%', cisloPoslednihoStavu + 2);

                            aktualniText = "";
                            stav = 8;
                        }
                        else if (z == '$')
                        {
                            int cisloPoslednihoStavu = this.AutomatVyrazu.Stavy.Count - 1;
                            //hrany do následujícího nového stavu
                            this.AutomatVyrazu.PridejStav();
                            foreach (char znak in aktualniText)
                            {
                                this.AutomatVyrazu.Stavy[cisloPoslednihoStavu].PridejHranu(
                                    Convert.ToChar(znak), cisloPoslednihoStavu+1);
                            }

                            aktualniText = "";
                            this.DefinovanKonec = true;
                            stav = 9;
                        }
                        else
                        {
                            validniVyraz = false;
                        }

                        break;
                    default:
                        validniVyraz = false;
                        break;
                }
            }

            if (aktualniText.Length > 0)
            {
                int cisloPoslednihoStavu = this.AutomatVyrazu.Stavy.Count - 1;
                this.AutomatVyrazu.PridejStav();
                foreach (char znak in aktualniText)
                {
                    this.AutomatVyrazu.Stavy[cisloPoslednihoStavu].PridejHranu(
                        Convert.ToChar(znak), cisloPoslednihoStavu + 1);
                }
            }

            //kontrola koncových stavů
            if (!(stav == 1 || stav == 6 || stav == 7 || stav == 8 || stav == 9 || stav == 11 || stav == 12))
            {
                validniVyraz = false;
            }

            //poslední stav je konečný
            this.AutomatVyrazu.Stavy[this.AutomatVyrazu.Stavy.Count - 1].JeKonecny();

            this.VyrazOK = validniVyraz;

        }

        /// <summary>
        /// Zjistí zda je zadané číslo z daného intervalu
        /// </summary>
        /// <param name="cislo">Číslo u něhož chceme zjistit zda se nachází v intervalu</param>
        /// <param name="zacatekIntervalu">Začátek intervalu (včetně)</param>
        /// <param name="konecIntervalu">Konec intervalu (včetně)</param>
        /// <returns></returns>
        private bool JeCisloZIntervalu(int cislo, int zacatekIntervalu, int konecIntervalu)
        {

            if (cislo >= zacatekIntervalu && cislo <= konecIntervalu)
            {
                return true;
            }
            else
            {
                return false;
            }
        }

    }

}
